from abc import ABC, abstractmethod
from base64 import b64encode
from typing import Dict

from click import Abort

from delfino_core.config import IssueTrackingConfig

try:
    import httpx
except ImportError:
    pass


class _BaseIssuerTrackerClient(ABC):
    def __init__(self, settings: IssueTrackingConfig):
        self._settings = settings

    @abstractmethod
    def get_issue_title(self, issue_number: int) -> str:
        """Fetches the title of an issue using its issue number."""


class JiraClient(_BaseIssuerTrackerClient):
    def _headers(self) -> Dict[str, str]:
        token = b64encode(f"{self._settings.username}:{self._settings.api_key}".encode()).decode()
        return {
            "Accept": "application/json",
            "Authorization": f"Basic {token}",
        }

    def get_issue_title(self, issue_number: int) -> str:
        issue_key = f"{self._settings.issue_prefix.rstrip('-')}-{issue_number}"
        url = f"{self._settings.tracker_url.rstrip('/')}/rest/api/3/issue/{issue_key}"

        response = httpx.get(url, headers=self._headers())

        if response.status_code == 200:
            data = response.json()
            return data["fields"]["summary"]

        raise Abort(f"Failed to fetch issue: {response.status_code}, {response.text}")



from __future__ import annotations

import logging
import re
from functools import lru_cache
from subprocess import PIPE
from typing import Literal

from delfino import run
from delfino.execution import OnError
from delfino.validation import assert_pip_package_installed

from delfino_core.commands.issue_tracker import JiraClient
from delfino_core.config import VCSConfig
from delfino_core.utils import ask, assert_executable_installed


@lru_cache(maxsize=1)
def get_trunk_branch():
    """To dynamically figure out if it's `main`, `master` or something else."""
    ref = run(["git", "symbolic-ref", "refs/remotes/origin/HEAD"], stdout=PIPE, on_error=OnError.PASS)

    # The ref is expected to be like 'refs/remotes/origin/main'
    return ref.stdout.decode().strip().split("/")[-1]


_LOGGER = logging.getLogger(__name__)
_INVALID_BRANCH_NAME_CHARS = re.compile(r"[^a-zA-Z0-9_/-]+")


def consume_args_until_next_option(passed_args: list[str]) -> tuple[str, list[str]]:
    for index, arg in enumerate(passed_args):
        if arg.startswith("-"):
            return " ".join(passed_args[:index]), passed_args[index:]
    return " ".join(passed_args), []


def _sanitize_branch_name(branch_name: str) -> str:
    return "/".join(_INVALID_BRANCH_NAME_CHARS.sub("_", part).strip("_") for part in branch_name.lower().split("/"))


def _get_user_name() -> str:
    # Get the current username from git config
    username = run(["git", "config", "user.email"], stdout=PIPE, on_error=OnError.ABORT).stdout.decode().strip()

    if not username:  # If not available, use the system username
        _LOGGER.warning("No git user.email found, using system username.")
        assert_executable_installed("whoami", required_by="gh")
        username = run(["whoami"], stdout=PIPE, on_error=OnError.ABORT).stdout.decode().strip()
    elif "@" in username:  # Strip the domain from the email address
        username = username.split("@")[0]

    return username


def get_new_branch_name_or_switch_to_branch(branch_prefix: str | None, title: str) -> tuple[str, bool]:
    if branch_prefix is None:
        branch_prefix = _get_user_name()

    if branch_prefix != "":  # allow for no prefix
        branch_prefix = _sanitize_branch_name(branch_prefix) + "/"

    branch_name = branch_prefix + _sanitize_branch_name(title)

    # Check if not already on the branch
    branch_exists = (
        run(["git", "branch", "--list", branch_name], stdout=PIPE, on_error=OnError.PASS).stdout.decode().strip()
    )

    if branch_exists:
        if branch_exists.startswith("* "):
            _LOGGER.warning(f"Already on branch '{branch_name}'.")
        else:
            _LOGGER.warning(f"Branch '{branch_name}' already exists, switching to it.")
            run(["git", "checkout", branch_name], on_error=OnError.ABORT)
    else:
        # Ask user if they want to use the branch name or let them choose a different one
        while not ask(f"Use branch name '{branch_name}'?"):
            new_branch_name = input(f"Enter a branch name: {branch_prefix}")
            sanitized_new_branch_name = _sanitize_branch_name(new_branch_name)
            if sanitized_new_branch_name == new_branch_name:  # not sanitized, use it
                branch_name = branch_prefix + new_branch_name
                break

    return branch_name, not branch_exists


@lru_cache(maxsize=1)
def get_vcs_cli_tool() -> Literal["gh", "glab"]:
    """Determine if `gh` or `glab` should be used."""
    remote_url = run(["git", "remote", "-v"], stdout=PIPE, on_error=OnError.PASS).stdout.decode()

    if "github" in remote_url:
        return "gh"

    if "gitlab" in remote_url:
        return "glab"

    raise RuntimeError("No supported VCS found in the git remote.")


def title_and_branch_prefix_from_issue_tracker(
    vcs_cli_tool: str, title: str, command_config: VCSConfig
) -> tuple[str, str | None]:
    branch_prefix = command_config.branch_prefix
    prompt_part = ""

    if command_config.issue_tracking.tracker_url:
        assert_pip_package_installed("httpx")
        prompt_part = " or issue number"

    while not title:
        title = input(f"Enter a title{prompt_part} for the {'PR' if vcs_cli_tool == 'gh' else 'MR'}: ").strip()

    if not command_config.issue_tracking.tracker_url:
        return title, branch_prefix

    issue_number = None
    try:
        issue_number = int(title)
    except ValueError:
        pass

    if issue_number is not None:
        title = JiraClient(command_config.issue_tracking).get_issue_title(issue_number)
        branch_prefix = f"{command_config.issue_tracking.issue_prefix}{issue_number}"

    return title, branch_prefix



from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

import click
from click import Abort
from delfino.decorators import files_folders_option, pass_args
from delfino.execution import OnError, run
from delfino.terminal_output import print_header
from delfino.validation import assert_pip_package_installed

try:
    import yaml
except ImportError:
    pass


def _selected_stages_and_hook(passed_args: List[str]) -> Tuple[List[str], Optional[str]]:
    pre_commit_file = Path(".pre-commit-config.yaml")
    if not pre_commit_file.is_file():
        raise Abort(f"Pre-commit config file '{pre_commit_file}' not found.")

    pre_commit_config = yaml.safe_load(pre_commit_file.read_bytes())

    default_stages = pre_commit_config.get("default_stages", [])
    all_stages: Set[str] = set(default_stages)
    hooks_to_stages: Dict[str, List[str]] = {}

    for repo in pre_commit_config["repos"]:
        for hook in repo["hooks"]:
            all_stages.update(hook.get("stages", []))
            hooks_to_stages[hook.get("name", hook["id"])] = hook.get("stages", default_stages)

    if len(passed_args) >= 1:
        hook_name = passed_args[0]
        if stages := hooks_to_stages.get(hook_name):
            # User selected a single tool
            return [stages[0]], hook_name

    return sorted(all_stages), None


@click.command("pre-commit")
@pass_args
@files_folders_option
@click.option(
    "--add",
    "-a",
    "stage_all_files",
    is_flag=True,
    default=False,
    help="Stage all files before running pre-commit hooks.",
)
def run_pre_commit(stage_all_files: bool, files_folders: List[Path], passed_args: List[str]):
    """Run all pre-commit stages in the current project (alias for `pre-commit run ...`).

    To run a single hook, add the name of the hook at the end, as if you were running
    `pre-commit run <HOOK NAME>`.
    """
    assert_pip_package_installed("PyYAML")

    if stage_all_files:
        run(["git", "add", "."], on_error=OnError.PASS)

    files = ["--files", *files_folders] if files_folders else []

    stages, hook = _selected_stages_and_hook(passed_args)

    if hook is None:
        msg = "all pre-commit stages" if len(stages) > 1 else "pre-commit"
        print_header(f"Running {msg}")

    for stage in stages:
        if len(stages) > 1:
            print_header(f"{stage} / {hook}" if hook else stage, level=2)

        run(
            [
                "pre-commit",
                "run",
                "--hook-stage",
                stage,
                *([hook] if hook else []),
                *passed_args,
                *files,
            ],
            on_error=OnError.PASS,
        )



from __future__ import annotations

import os
import re
import webbrowser
from datetime import datetime, timedelta
from pathlib import Path
from subprocess import PIPE, CompletedProcess

import click
from click import secho
from delfino.constants import ENTRY_POINT, PackageManager
from delfino.decorators import pass_app_context
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header
from delfino.validation import assert_package_manager_is_known, assert_pip_package_installed

from delfino_core.commands.verify import run_group_verify
from delfino_core.config import CorePluginConfig
from delfino_core.spinner import Spinner
from delfino_core.utils import ask

try:
    from git import Repo
except ImportError:
    pass

try:
    import yaml
except ImportError:
    pass


def _run(args: str, spinner: Spinner | None = None) -> CompletedProcess:
    """Print the command before execution."""
    if spinner is None:
        return run(args, on_error=OnError.EXIT, stdout=PIPE, stderr=PIPE)

    result = run(args, on_error=OnError.PASS, running_hook=spinner, stdout=PIPE, stderr=PIPE)
    spinner.print_results(result, error_cls=click.exceptions.Exit)
    return result


class Changelogs:
    _DATA_FILE = Path(__file__).parent.parent / "changelog_urls.yaml"

    def __init__(self):
        assert_pip_package_installed("PyYAML")

        with open(self._DATA_FILE, encoding="utf-8") as file:
            self._packages_to_urls: dict[str, str] = yaml.safe_load(file)["changelogs"]

    def url_for_package(self, package: str) -> str:
        return self._packages_to_urls.get(package, "")


class Updater:
    _FILENAME: str = ""

    @staticmethod
    def _git_root():
        return (
            run(["git", "rev-parse", "--show-toplevel"], stdout=PIPE, stderr=PIPE, on_error=OnError.EXIT)
            .stdout.decode()
            .strip()
        )

    def __init__(self, stash: bool):
        assert_pip_package_installed("gitpython")

        self._repo = Repo(self._git_root())
        self._stash = stash
        now = datetime.utcnow()
        self._start_of_week = now - timedelta(now.isoweekday() - 1)
        self._changelog = Changelogs()

    def commit_and_push(self):
        commit_message = f"Dependencies rollup: {self._start_of_week.strftime('%Y-%m-%d')}"
        can_update = self._repo.commit("HEAD").message.strip() == commit_message

        if do_commit := ask("Do you want to commit changes now?"):
            if can_update:  # update existing commit
                if self._repo.is_dirty():
                    _run("git pull")
                _run("git commit -a --amend -C HEAD")
            else:
                _run("git add .")
                _run(f"git commit -a -m '{commit_message}'")

        if do_commit and ask("Do you want to push changes now?"):
            if can_update:
                _run("git push --force-with-lease")
            else:
                _run("git push -u origin HEAD")

            url = self._link_to_open_a_pull_request()
            if url:
                if ask("Do you want to open a new pull request now in a web browser?"):
                    webbrowser.open(url)
                else:
                    secho(f"\nOpen a new pull request by visiting:\n\n\t{url}\n", fg="green")

    def _link_to_open_a_pull_request(self) -> str | None:
        url = self._repo.remote().url

        if not (match := re.match("git@github.com:(.*)\\.git", url)):
            match = re.match("https://github.com/(.*)\\.git", url)

        return f"https://github.com/{match.group(1)}/pull/new/{self._repo.active_branch}" if match else None

    def _read_dependency_file(self) -> str:
        with open(self._FILENAME, encoding="utf-8") as file:
            return file.read()

    def print_outdated_packages_and_lock_if_changed(self) -> bool:
        raise NotImplementedError

    def get_branch_name(self) -> str:
        email = self._repo.config_reader().get_value("user", "email")
        assert (
            email
        ), "Git config has not a commit email set. Please set it with:\n\tgit config --global user.email 'YOUR_EMAIL'"
        assert isinstance(email, str)

        user_name = re.sub("@.*", "", email)
        assert user_name, f"No user name could be parsed from git commit email '{email}' after removing '@.*'."
        assert isinstance(user_name, str)
        user_name = re.sub("[^a-zA-Z0-9._-]", "_", user_name)

        return f"{user_name}/dependencies_rollup_{self._start_of_week.strftime('%Y_%m_%d')}"

    @classmethod
    def _lock_and_sync(cls):
        pass

    def _show_edit_prompt_and_wait(self, *, available_updates: str):
        print("\n" + available_updates)
        input(
            f"\033[1;33mEdit {self._FILENAME} file to update any of the dependencies above ^^^\n\n"
            f"Then continue by pressing ENTER ...\033[0m"
        )

    def checkout_branch(self, branch: str):
        if str(self._repo.active_branch) == branch:
            secho(f"Branch '{branch}' already exists and active.", fg="green")
        elif branch in self._repo.branches:  # type: ignore[operator]
            secho(f"Branch '{branch}' already exists.", fg="yellow")
            _run(f"git checkout {branch}")
        else:
            if self._stash:
                secho("Stashing existing changes.", fg="yellow")
                _run("git stash")

            if self._repo.active_branch != "main":
                secho("Checking out 'main'.", fg="yellow")
                _run("git checkout main")

            secho("Pulling latest changes.", fg="yellow")
            _run("git pull")

            secho(f"Creating a new branch '{branch}'.", fg="yellow")
            _run(f"git checkout -b {branch}")
            self._lock_and_sync()

    def update(self, retry: bool, create_branch: bool):
        if create_branch:
            branch_name = self.get_branch_name()
            self.checkout_branch(branch_name)

        if not retry:
            while True:
                if not self.print_outdated_packages_and_lock_if_changed():
                    break

            secho("Running all tests to check the updated dependencies ... ", fg="yellow")


class PipenvUpdater(Updater):
    _FILENAME = "Pipfile"

    _SKIP_PATTERN = (
        "Skipped Update of Package (?P<package>[^:]+): (?P<installed>[^ ]+) "
        "installed.* (?P<available>[^ ]+) available.*"
    )
    _OUTDATED_PATTEN = (
        "Package '(?P<package>[^']+)' out-of-date: '(?P<installed>[^']+)' "
        "installed.* '(?P<available>[^']+)' available.*"
    )
    _VERSION_CONSTRAINT_CHARS = "=~<>"

    @classmethod
    def _lock_and_sync(cls):
        _run("pipenv lock")
        _run("pipenv sync -d")

    def print_outdated_packages_and_lock_if_changed(self) -> bool:
        spinner = Spinner("pipenv", "updating packages based on version pinning")
        _run("pipenv update -d", spinner)

        spinner = Spinner("pipenv", "checking outdated packages")
        result = _run("pipenv update --outdated", spinner)

        pipfile = self._read_dependency_file()

        available_updates = []

        for line in result.stdout.decode().split(os.linesep) + result.stderr.decode().split(os.linesep):
            if not (match := re.match(self._SKIP_PATTERN, line) or re.match(self._OUTDATED_PATTEN, line)):
                continue

            package = match.group("package")
            installed = match.group("installed").lstrip(self._VERSION_CONSTRAINT_CHARS)
            available = match.group("available").lstrip(self._VERSION_CONSTRAINT_CHARS)

            # Keep only packages defined in Pipfile with a different version available
            if installed != available and re.search(f'\n"?{package}"? ', pipfile):
                if changelog_url := self._changelog.url_for_package(package):
                    changelog_url = f" ({changelog_url})"
                available_updates.append(f"{package}: {installed} -> {available}{changelog_url}")

        if not available_updates:
            return False

        self._show_edit_prompt_and_wait(available_updates="\n".join(sorted(available_updates)) + "\n")

        return self._read_dependency_file() != pipfile


class PoetryUpdater(Updater):
    _FILENAME = "pyproject.toml"
    _ANSI_ESCAPE = re.compile(r"\x1b\[[0-9;]*m")

    def parse_package_name(self, line: str) -> str:
        return self._ANSI_ESCAPE.sub("", line).split(" ", maxsplit=1)[0]

    def print_outdated_packages_and_lock_if_changed(self) -> bool:
        spinner = Spinner("poetry", "updating packages based on version pinning")
        _run("poetry update", spinner)

        spinner = Spinner("poetry", "checking outdated packages")
        if not (result := _run("poetry show --outdated --why --ansi", spinner).stdout.decode()):
            return False

        pyproject_toml = self._read_dependency_file()
        updates = []

        for line in result.split(os.linesep):
            if not line:
                continue

            package = self.parse_package_name(line)
            if changelog_url := self._changelog.url_for_package(package):
                changelog_url = f" ({changelog_url})"

            updates.append(f"{line}{changelog_url}")

        self._show_edit_prompt_and_wait(available_updates="\n".join(updates))

        return self._read_dependency_file() != pyproject_toml


@click.command("dependencies-update")
@click.option("--retry", default=False, show_default=True, is_flag=True, help="Retry an update after failed tests.")
@click.option(
    "--stash/--no-stash",
    default=True,
    show_default=True,
    help=(
        "Don't stash existing changes. Useful when there are some existing manual changes "
        "that should be also part of the upgrade."
    ),
)
@click.option("--no-branch", default=False, show_default=True, is_flag=True, help="Don't create a new branch.")
@pass_app_context(CorePluginConfig)
@click.pass_context
def run_dependencies_update(
    click_context: click.Context, app_context: AppContext, retry: bool, stash: bool, no_branch: bool
):
    """Manages the process of updating dependencies."""
    print_header("Updating dependencies", icon="🔄")

    assert_package_manager_is_known(app_context.package_manager)
    updater: Updater

    if app_context.package_manager == PackageManager.PIPENV:
        updater = PipenvUpdater(stash)
    elif app_context.package_manager == PackageManager.POETRY:
        updater = PoetryUpdater(stash)
    else:
        raise AssertionError(
            f"The '{app_context.package_manager.value}' package manager is not supported by this command."
        )

    updater.update(retry, create_branch=not no_branch)

    try:
        click_context.invoke(run_group_verify)
    except Exception:
        secho(
            f"\nOne or more checks have failed. Fix any issues above and then run:\n\n\t"
            f"{ENTRY_POINT} {click_context.info_name} --retry\n",
            fg="red",
        )
        raise

    updater.commit_and_push()



from collections import ChainMap
from logging import getLogger
from subprocess import PIPE, run
from typing import Dict, cast

import click
from delfino.click_utils.command import get_root_command
from delfino.decorators.files_folders import FILES_FOLDERS_OPTION_CALLBACK
from delfino.decorators.pass_args import PASS_ARGS_CALLBACK

from delfino_core.config import CorePluginConfig

_LOG = getLogger(__name__)


def ensure_reports_dir(config: CorePluginConfig) -> None:
    """Ensures that the reports directory exists."""
    config.reports_directory.mkdir(parents=True, exist_ok=True)


def commands_group_help(name: str) -> str:
    command_names = ", ".join(CorePluginConfig.model_fields[f"{name}_commands"].default)
    return f"Runs {command_names}.\n\n" f"Configured by the ``{name}_commands`` settings option."


def execute_commands_group(click_context: click.Context, plugin_config: CorePluginConfig, **kwargs):
    """Executes a group of commands.

    Group default commands are defined in `delfino_core.config.CorePluginConfig.<NAME>_commands`.
    They can be overridden by `<NAME>_commands` option in the `pyproject.toml` config.
    """
    root = get_root_command(click_context)
    if (name := click_context.command.name) is None:
        raise ValueError("group commands must have a name set")

    option_name = f"{name.replace('-', '_')}_commands"

    commands: Dict[str, click.Command] = {
        command: cast(click.Command, root.get_command(click_context, command))
        for command in root.list_commands(click_context)
    }

    for target_name in getattr(plugin_config, option_name):
        if target_name not in commands:
            _LOG.warning(
                f"Command '{target_name}' used as part of the '{option_name}' option does not exist. Skipping."
            )
            continue

        if target_name in plugin_config.disable_commands:
            _LOG.debug(f"Skipping disabled command '{target_name}'.")
            continue

        command = commands[target_name]

        parameter_from_config = ChainMap(
            *(
                callback.parameter_from_config_in_group(click_context, command)
                for callback in [PASS_ARGS_CALLBACK, FILES_FOLDERS_OPTION_CALLBACK]
            )
        )

        click_context.forward(command, **kwargs, **parameter_from_config)


def executable_installed(name: str, *flags: str) -> bool:
    """Returns True if the executable is installed."""
    if not flags:
        flags = ("--version",)

    try:
        run([name, *flags], stdout=PIPE, check=True)
    except FileNotFoundError:
        return False
    return True


def assert_executable_installed(name: str, *flags: str, required_by: str = "this command") -> None:
    assert executable_installed(
        name, *flags
    ), f"Optional executable '{name}' is required by {required_by} but not installed."


def ask(question: str) -> bool:
    return not bool(input(f"\033[1;33m{question} [Y/n]: \033[0m").lower() == "n")
